match self {
FileType::CL_TYPE_7Z => Some(Feature::FileType7Z),
FileType::CL_TYPE_7ZSFX => Some(Feature::FileType7Zsfx),
FileType::CL_TYPE_AI_MODEL => Some(Feature::FileTypeAiModel),
FileType::CL_TYPE_ALZ => Some(Feature::FileTypeAlz),
FileType::CL_TYPE_ANY => Some(Feature::FileTypeAny),
FileType::CL_TYPE_APM => Some(Feature::FileTypeApm),
FileType::CL_TYPE_CPIO_CRC => Some(Feature::FileTypeCpioCrc),
FileType::CL_TYPE_CPIO_NEWC => Some(Feature::FileTypeCpioNewc),
FileType::CL_TYPE_CPIO_ODC => Some(Feature::FileTypeCpioOdc),
FileType::CL_TYPE_CPIO_OLD => Some(Feature::FileTypeCpioOld),
FileType::CL_TYPE_DMG => Some(Feature::FileTypeDmg),
FileType::CL_TYPE_EGG => Some(Feature::FileTypeEgg),
FileType::CL_TYPE_EGGSFX => Some(Feature::FileTypeEggsfx),
FileType::CL_TYPE_EPS => Some(Feature::FileTypeEps),
FileType::CL_TYPE_GIF => Some(Feature::FileTypeGif),
FileType::CL_TYPE_GPT => Some(Feature::FileTypeGpt),
FileType::CL_TYPE_HWP3 => Some(Feature::FileTypeHwp3),
FileType::CL_TYPE_HWPOLE2 => Some(Feature::FileTypeHwpole2),
FileType::CL_TYPE_INTERNAL => Some(Feature::FileTypeInternal),
FileType::CL_TYPE_ISHIELD_MSI => Some(Feature::FileTypeIshieldMsi),
FileType::CL_TYPE_ISO9660 => Some(Feature::FileTypeIso9660),
FileType::CL_TYPE_JAVA => Some(Feature::FileTypeJava),
FileType::CL_TYPE_JPEG => Some(Feature::FileTypeJpeg),
FileType::CL_TYPE_LHA_LZH => Some(Feature::FileTypeLhaLzh),
FileType::CL_TYPE_LNK => Some(Feature::FileTypeLnk),
FileType::CL_TYPE_MACHO => Some(Feature::FileTypeMacho),
FileType::CL_TYPE_MACHO_UNIBIN => Some(Feature::FileTypeMachoUnibin),
FileType::CL_TYPE_MBR => Some(Feature::FileTypeMbr),
FileType::CL_TYPE_MHTML => Some(Feature::FileTypeMhtml),
FileType::CL_TYPE_ONENOTE => Some(Feature::FileTypeOnenote),
FileType::CL_TYPE_OOXML_HWP => Some(Feature::FileTypeOoxmlHwp),
FileType::CL_TYPE_OOXML_PPT => Some(Feature::FileTypeOoxmlPpt),
FileType::CL_TYPE_OOXML_WORD => Some(Feature::FileTypeOoxmlWord),
FileType::CL_TYPE_OOXML_XL => Some(Feature::FileTypeOoxmlXl),
FileType::CL_TYPE_OTHER => Some(Feature::FileTypeOther),
FileType::CL_TYPE_PART_ANY => Some(Feature::FileTypePartAny),
FileType::CL_TYPE_PART_HFSPLUS => Some(Feature::FileTypePartHfsplus),
FileType::CL_TYPE_PNG => Some(Feature::FileTypePng),
FileType::CL_TYPE_PS => Some(Feature::FileTypePs),
FileType::CL_TYPE_PYTHON_COMPILED => Some(Feature::FileTypePythonCompiled),
FileType::CL_TYPE_SWF => Some(Feature::FileTypeSwf),
FileType::CL_TYPE_TIFF => Some(Feature::FileTypeTiff),
FileType::CL_TYPE_UDF => Some(Feature::FileTypeUdf),
FileType::CL_TYPE_XAR => Some(Feature::FileTypeXar),
FileType::CL_TYPE_XDP => Some(Feature::FileTypeXdp),
FileType::CL_TYPE_XML_HWP => Some(Feature::FileTypeXmlHwp),
FileType::CL_TYPE_XML_WORD => Some(Feature::FileTypeXmlWord),
FileType::CL_TYPE_XML_XL => Some(Feature::FileTypeXmlXl),
FileType::CL_TYPE_XZ => Some(Feature::FileTypeXz),
_ => None,
}
